import logging
import concurrent.futures
import pandas
import yaml

from .             import DataCache, MarketIndex, CurrencyExchange, PortfolioAggregator
from .currency     import brasil_banco_central     as currency_bcb
from .currency     import cryptocompare            as currency_cryptocompare
from .marketindex  import brasil_banco_central     as mktidx_bcb
from .marketindex  import federal_reserve          as mktidx_fred
from .marketindex  import yahoo_finance            as mktidx_yahoo
from .portfolios   import google_sheets            as google_sheets
from .portfolios   import uri                      as uri







class Investor(object):
    domains={'portfolio','currency_converters','benchmarks'}



    def __init__(self,file,refreshMap=dict(zip(domains,len(domains)*[False])),load=True):
        # Setup logging
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        defaultRefreshMap=dict(zip(self.domains,len(self.domains)*[False]))

        refresh=defaultRefreshMap
        refresh.update(refreshMap)

        with open(file, 'r', encoding="utf8") as f:
            self.config = yaml.load(f, Loader=yaml.FullLoader)

        self.currency=self.config['currency']

        self._cache               = None
        self.portfolio            = None
        self.currency_converters  = None
        self.benchmarks           = None
        self.exchange             = CurrencyExchange(self.currency)

        if load:
            # Load all time series data from cache or web and compute derived
            # data
            self.loadDomains(refresh)



    def loadDomains(self,refreshMap=dict(zip(domains,len(domains)*[False]))):
        """
        Load from cache or internet data for Portfolio, Currency Converters and
        Benchmarks.

        refreshMap is a dict that controls wether to load from cache or update
        domain from internet (and update cache). It looks like:

        dict(
            portfolio           = False,
            benchmarks          = False,
            currency_converters = False,
        )
        """

        defaultRefreshMap=dict(zip(self.domains,len(self.domains)*[False]))

        refresh=defaultRefreshMap
        refresh.update(refreshMap)

        # Decide wether we need to update extra benchmarks based on
        # CurrencyConverters
        augmentDomains=False
        for dom in ['benchmarks','currency_converters']:
            augmentDomains=augmentDomains or refresh[dom] or not getattr(self,dom)

        # Decide wether we need to update CurrencyExchange object
        updateCurrencyExchange=False
        for dom in ['currency_converters']:
            updateCurrencyExchange=(
                updateCurrencyExchange or
                refresh[dom] or not
                getattr(self,dom)
            )

        with concurrent.futures.ThreadPoolExecutor(thread_name_prefix='load_domains') as executor:
            tasks={}

            for domain in self.domains:
                if refresh[domain] or getattr(self,domain) is None:
                    # Load if a refresh was requested or domain has nothing yet
                    setattr(self,domain,[])
                    for part in self.config[domain]:
                        if 'type' in part:
#                             self.logger.info(f"processing {part['type']}")
#                             self.logger.info(f"{part['params'].copy()}")
                            # If it contains a class that needs activation or loading

                            # Prepare parameters
                            theparams=part['params'].copy()
                            theparams.update(
                                dict(
                                    cache   = self.cache,
                                    refresh = refresh[domain]
                                )
                            )


                            task=executor.submit(
                                # The class
                                part['type'],

                                # The parameters for class’ __init__()
                                **theparams
                            )
                            tasks[task]=(domain,part)

            for task in concurrent.futures.as_completed(tasks):
                # Unpack what we’ve set previously
                (domain,part)=tasks[task]

                # Put resulting object under its config
                part['obj']=task.result()

                # Add to internal inventory
                getattr(self,domain).append(part)

        # Add more benchmarks as per config file
        if augmentDomains:
            self.augmentDomains()

        # Rearrange Portfolio
        if len(self.portfolio)>1:
            agg=PortfolioAggregator()
            agg.append([p['obj'] for p in self.portfolio])
            self.portfolio=agg
        else:
            self.portfolio=self.portfolio[0]['obj']

        if updateCurrencyExchange:
            # Setup a multiple currency exchange machine
            self.exchange=CurrencyExchange(self.currency)

            # Put all CurrencyConverters in a single useful CurrencyExchange machine
            for curr in self.currency_converters:
                self.exchange.addCurrency(curr['obj'])


    def augmentDomains(self):
        """
        Fabricate more benchmarks from currency converters, as specified by
        YAML config file.

        Entries such as the following from the YAML file will be engineered:

        benchmarks:
            - kind: from_currency_converter
              from_to: BRLUSD
        """

        for item in self.config['benchmarks']:
            if 'kind' in item and item['kind'] == 'from_currency_converter':
                curFrom = item['from_to'][:3]
                curTo   = item['from_to'][3:]

                benchmark_signature = '[{currency}] {name}'.format(
                    name=item['from_to'],
                    currency=curTo
                )

                # Delete old MarketIndex previously created by a CurrencyConverter
                self.benchmarks = [b for b in self.benchmarks if b['obj'].id!=curFrom+curTo]

                # Scan all currency converters we have to find a match
                for cc in self.currency_converters:
                    if curFrom == cc['obj'].currencyFrom and curTo == cc['obj'].currencyTo:
                        item['obj']=MarketIndex().fromCurrencyConverter(cc['obj'])
                        break
                    elif curTo == cc['obj'].currencyFrom and curFrom == cc['obj'].currencyTo:
                        item['obj']=MarketIndex().fromCurrencyConverter(cc['obj'].invert())
                        break

                if 'obj' in item:
                    self.benchmarks.append(item)
                else:
                    raise Exception(f'Can’t find "{item.kind}" CurrencyConverter to make a MarketIndex from.')

        # Sort benchmarks
        benchs={str(b['obj']): b for b in self.benchmarks}
        self.benchmarks=[benchs[k] for k in sorted(benchs)]



    @property
    def cache(self):
        if self._cache is None:
            self._cache=DataCache(self.config['cache_database'],recycle=2)
            self.logger.debug(f"Using cache: {self._cache}")

        return self._cache



    @property
    def currency(self):
        return self._currency



    @currency.setter
    def currency(self,name):
        self._currency=name

        if hasattr(self,'exchange'):
            self.exchange.currency=name



    def __repr__(self):
        return yaml.dump(
            {
                'Portfolio': self.portfolio.__repr__(),
                'Exchange': self.exchange.__repr__(),
                'Benchmarks': [x.__repr__() for x in self.benchmarks],
                'Currency converters': [x.__repr__() for x in self.currency_converters],
            },
            indent=6,
            canonical=False,
            default_flow_style=False
        )



import datetime
import logging
import urllib
import requests
import numpy
import pandas

from .. import CurrencyConverter


class BCBCurrencyConverter(CurrencyConverter):
    """
    CurrencyConverter with data from Banco Central do Brasil.
    currencyFrom supports 'USD' and 'EUR'.

    https://dadosabertos.bcb.gov.br/dataset/dolar-americano-usd-todos-os-boletins-diarios

    https://dadosabertos.bcb.gov.br/dataset/taxas-de-cambio-todos-os-boletins-diarios
    """
    def __init__(self, currencyFrom, cache=None, refresh=False):
        super().__init__(
            currencyFrom  = currencyFrom,
            currencyTo    = 'BRL',

            kind          = 'BCBCurrencyConverter',
            id            = currencyFrom,

            cache         = cache,
            refresh       = refresh
        )



    def refreshData(self):
        # Format for start in BCB API parameters is MM-DD-YYYY.
        tomorrow=(datetime.date.today()+datetime.timedelta(days=1)).strftime('%m-%d-%Y')

        start='01-01-1970'
        if self.currencyFrom=='USD':
            # The of birth of BRLUSD. Before this date fiat was another thing and values can't be mixed
            start='07-01-1994' # MM-DD-YYYY

        ptax="https://olinda.bcb.gov.br/olinda/servico/PTAX/versao/v1/odata/CotacaoMoedaPeriodo(moeda=@moeda,dataInicial=@dataInicial,dataFinalCotacao=@dataFinalCotacao)"

        ptaxParams={
            '@moeda':                f"'{self.currencyFrom}'",
            '@dataInicial':          f"'{start}'",
            '@dataFinalCotacao':     f"'{tomorrow}'",
            '$select':               'dataHoraCotacao,cotacaoCompra,cotacaoVenda',
            '$format':               'json',
        }

        ptaxParamsStr = urllib.parse.urlencode(ptaxParams, safe="@$', ")

        response=requests.get(ptax,params=ptaxParamsStr)

        self.data=pandas.DataFrame(response.json()['value'])

        self.data.rename(columns={'dataHoraCotacao': 'time'}, inplace=True)



    def processData(self):
        self.data = (
            self.data

            .assign(
                value = lambda table: (table.cotacaoCompra+table.cotacaoVenda)/2,
                std   = lambda table: abs(table.cotacaoCompra/table.cotacaoVenda-1)
            )

            # Remove impossible values
            .query("cotacaoCompra!=0 and cotacaoVenda!=0 and std<0.2")

            .drop('cotacaoCompra cotacaoVenda std'.split(),axis=1)

            .assign(
                time = lambda table: (
                    pandas.to_datetime(table.time)

                    # Data source time is delivered as Brasilia local time
                    .dt.tz_localize('Brazil/East')

                    # Keep it as UTC internally because this is our standard
                    .dt.tz_convert('UTC')
                )
            )

            .drop_duplicates()

            .set_index('time')

            .sort_index()
        )



import logging
import sqlalchemy
import pandas


# TODO: Convert all queries to SQLAlchemy methods


class DataCache(object):
    """
    Implements a simple generic cache in a local SQLite database.

    Time series for market indices, currency converters and even your portfolio
    kept on Google Sheets take a lot of time to load because they are published
    as slow web APIs. To speed up initial data loading, all classes in the
    investzilla framework know how to work with a DataCache.

    A dataset that needs to be cached can have arbitrary columns and is usually
    the raw data as returned by the web API, after column names normalization
    but before cleanup and processing.

    A dataset has a `kind` and `ID`. So for example, the YahooMarketIndex class
    has kind `YahooMarketIndex` and the `ID` might be `^IXIC` or `^GSPC` -- the
    name of the market index you put in your portfolio configuration.

    DataCache will organize tables in the database with names
    `DataCache__{kind}`. And then each table will have columns:

    * __DataCache_id: values for our YahooMarketIndex examples will be `^IXIC`
      or `^GSPC`
    * __DataCache_time: UTC time the cache was set
    * other arbitrary columns found in the data attribute of set() method

    DataCache will keep multiple versions of the dataset, differentiated by
    __DataCache_time and it will always use the last version
    (max(__DataCache_time)). The maximum number of versions kept are defined by
    the __init__()’s recycle attribute. Older versions of data will be
    automatically deleted.
    """

    idCol       = '__DataCache_id'
    timeCol     = '__DataCache_time'
    typeTable   = 'DataCache__{kind}'



    def __init__(self, url='sqlite:///cache.db?check_same_thread=False', recycle=5):
        self.url=url
        self.db=None
        self.recycle=recycle

        # Setup logging
        self.getLogger()

        self.getDB()



    def __repr__(self):
        return 'DataCache(url={url},recycle={recycle})'.format(
            url=self.url,
            recycle=self.recycle
        )



    def __getstate__(self):
        o = self.__dict__
        o.update(
            dict(
                db = None,
                logger = None
            )
        )
        return o



    def getLogger(self):
        if hasattr(self,'logger')==False or self.logger is None:
            self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        return self.logger



    def getDB(self):
        if hasattr(self,'db') and self.db is not None:
            return self.db

        engine_config_sets=dict(
            # Documentation for all these SQLAlchemy pool control parameters:
            # https://docs.sqlalchemy.org/en/14/core/engines.html#engine-creation-api

            DEFAULT=dict(
                # QueuePool config for a real database
                poolclass         = sqlalchemy.pool.QueuePool,

                # 5 is the default. Reinforce default, which is good
                pool_size         = 5,

                # Default here was 10, which might be low sometimes, so
                # increase to some big number in order to never let the
                # QueuePool be a bottleneck.
                max_overflow      = 50,

                # Debug connection and all queries
                # echo              = True
            ),
            sqlite=dict(
                # SQLite doesn’t support concurrent writes, so we‘ll amend
                # the DEFAULT configuration to make the pool work with only
                # 1 simultaneous connection. Since Investorzilla is agressively
                # parallel and requires a DB service that can be used in
                # parallel (regular DBs), the simplicity and portability
                # offered by SQLite for a light developer laptop has its
                # tradeoffs and we’ll have to tweak it to make it usable in
                # a parallel environment even if SQLite is not parallel.

                # A pool_size of 1 allows only 1 simultaneous connection.
                pool_size         = 1,
                max_overflow      = 0,

                # Since we have only 1 stream of work (pool_size=1),
                # we need to put a hold on other DB requests that arrive
                # from other parallel tasks. We do this putting a high value
                # on pool_timeout, which controls the number of seconds to
                # wait before giving up on getting a connection from the
                # pool.
                pool_timeout      = 3600.0,

                # Debug connection and all queries
                # echo              = True
            ),
        )

        # Start with a default config
        engine_config=engine_config_sets['DEFAULT'].copy()

        # Add engine-specific configs
        for dbtype in engine_config_sets.keys():
            # Extract from engine_config_sets configuration specific
            # for each DB type
            if dbtype in self.url:
                engine_config.update(engine_config_sets[dbtype])

        if hasattr(self,'db')==False or self.db is None:
            self.getLogger().debug(f"Creating a DB engine on {self.url}")

            self.db=sqlalchemy.create_engine(
                url = self.url,
                **engine_config
            )

        return self.db



    def last(self, kind, id):
        """
        Return last time data was updated on cache for this kind and id
        """

        table=self.typeTable.format(kind=kind)

        q='''
            SELECT max({timeCol}) AS last
            FROM {typeTable}
            WHERE {idCol} = '{id}'
        '''

        query=q.format(
            typeTable     = table,
            idCol         = self.idCol,
            timeCol       = self.timeCol,
            id            = id
        )

        try:
            self.getLogger().debug(f'Trying cache as {query}')

            with self.getDB().connect() as db:
                df=pandas.read_sql(query,con=db)

            if df.shape[0]>0:
                self.logger.info(f"Successful cache hit for kind={kind} and id={id} with {df.shape[0]} records.")
                df['last']=pandas.to_datetime(df['last'])
                ret=df['last'][0]
                if ret.tzinfo is None or ret.tzinfo.utcoffset(ret) is None:
                    ret=ret=ret.tz_localize('UTC')
            else:
                self.getLogger().info(f"Cache empty for kind={kind} and id={id}")
                ret=None
        except Exception as e:
            self.getLogger().warning(f"No cache for kind={kind} and id={id}")
            self.getLogger().info(e)
            ret=None


        return ret



    def get(self, kind, id, time=None):
        """
        kind leads to table DataCache_{kind}

        id is written to column __DataCache_id

        time makes it search on column __DataCache_time for entries with id
        recent up to time.

        Returns a tuple with table and time of cache data.
        """

        table=self.typeTable.format(kind=kind)

        if time is None:
            pointInTime='''
            (
                SELECT DISTINCT {timeCol}
                FROM {typeTable}
                WHERE {idCol} = '{id}'
                ORDER BY {timeCol} DESC
                LIMIT 1
            )
            '''
        else:
            pointInTime='''
            (
                SELECT DISTINCT {timeCol}
                FROM {typeTable}
                WHERE
                    {idCol}    = '{id}' AND
                    {timeCol} <= '{time}'
                ORDER BY {timeCol} DESC
                LIMIT 1
            )
            '''

        pointInTime=pointInTime.format(
            timeCol      = self.timeCol,
            time         = time,
            idCol        = self.idCol,
            id           = id,
            typeTable    = table
        )

        query='''
            SELECT *
            FROM {typeTable}
            WHERE
                {idCol}   = '{id}' AND
                {timeCol} = {pointInTime}
        '''

        query=query.format(
            typeTable    = table,
            idCol        = self.idCol,
            timeCol      = self.timeCol,
            id           = id,
            pointInTime  = pointInTime
        )

        try:
            self.getLogger().debug(f'Trying cache as {query}')

            with self.getDB().connect() as db:
                df=pandas.read_sql(query,con=db)

            if df.shape[0]>0:
                age=pandas.Timestamp(df[self.timeCol].max())
                self.getLogger().info(f"Cache for kind={kind} and id={id} has {df.shape[0]} entries and was cached at {age}")
                return (df.drop(columns=[self.timeCol,self.idCol]),age)
            else:
                self.getLogger().info(f"Cache empty for kind={kind} and id={id}")
                return (None,None)

        except Exception as e:
            self.getLogger().info(f"No cache for kind={kind} and id={id}")
            self.getLogger().info(e)

            return (None,None)



    def cleanOld(self, kind, id):
        # Next query is overcomplicated because couldn’t make work its simplest version:
        # cleanQuery='''
        #     DELETE
        #     FROM {typeTable}
        #     WHERE {idCol} = '{id}'
        #     AND {timeCol} <= coalesce(
        #         (
        #             SELECT DISTINCT {timeCol}
        #             FROM {typeTable}
        #             WHERE {idCol} = '{id}'
        #             ORDER BY {timeCol} DESC
        #             limit 1 offset {recycle}
        #         ),
        #         date(0)
        #     )
        # '''

        # This query also doesn’t work flawlessly probably because of dead lock
        # https://stackoverflow.com/a/52467973/367824
        # cleanQuery='''
        #     WITH oldest AS (
        #         SELECT DISTINCT {timeCol}
        #         FROM {typeTable}
        #         WHERE {idCol} = '{id}'
        #         ORDER BY {timeCol} DESC limit 1 offset {recycle}
        #     )
        #     DELETE FROM {typeTable}
        #     WHERE EXISTS (
        #         SELECT 1
        #         FROM oldest
        #         WHERE {typeTable}.{timeCol} <= oldest.{timeCol}
        #     )
        # '''

        versionSelector='''
            SELECT DISTINCT {timeCol} AS deprecated
            FROM {typeTable}
            WHERE {idCol} = '{id}'
            ORDER BY {timeCol} DESC limit 1 offset {recycle}
        '''

        cleaner='''
            DELETE
            FROM {typeTable}
            WHERE {idCol} = '{id}'
            AND {timeCol} <= '{deprecated}'
        '''

        if self.recycle is not None:
            with self.getDB().connect() as db:
                deprecated=pandas.read_sql_query(
                    versionSelector.format(
                        typeTable    = self.typeTable.format(kind=kind),
                        idCol        = self.idCol,
                        timeCol      = self.timeCol,
                        id           = id,
                        recycle      = self.recycle
                    ),
                    con=db
                )

                if deprecated.shape[0]>0:
                    cleanQuery=cleaner.format(
                        typeTable    = self.typeTable.format(kind=kind),
                        idCol        = self.idCol,
                        timeCol      = self.timeCol,
                        id           = id,
                        deprecated   = deprecated.deprecated.iloc[0],
                        recycle      = self.recycle
                    )

                    self.getLogger().debug(f'Clean old cache entries as {cleanQuery}')
                    db.execute(sqlalchemy.text(cleanQuery))



    def set(self, kind, id, data):
        """
        kind leads to table DataCache__{kind}

        id is written to column __DataCache_id

        Current time is written to column __DataCache_time

        data is a DataFrame whose columns are the other columns of table DataCache__{kind}
        """

        d=data.copy()

        columns=list(d.columns)

        d[self.idCol]=id
        now=pandas.Timestamp.utcnow()
        d[self.timeCol]=now


        self.getLogger().info(f'Set cache to kind={kind}, id={id}, time={now}')

        with self.getDB().connect() as db:
            d[[self.idCol,self.timeCol] + columns].to_sql(
                self.typeTable.format(kind=kind),
                index       = False,
                if_exists   = 'append',
                chunksize   = 999,
                con         = db,
                method      = 'multi'
            )

        self.cleanOld(kind, id)


import os
import datetime
import pickle
import logging
import concurrent.futures
import numpy
import pandas


from . import Fund
from . import DataCache


class Portfolio(object):
    """
    A simple interface for a generic Portfolio.

    A Portfolio is the keeper of raw balance and ledger of multiple
    investment instruments, funds etc. Members of a Portfolio can be
    aggregated to generate one Fund which in turns has a time series for
    number of shares and a time series for share value.

    Derived classes must set self.ledger and self.balance in order for the
    Fund creation logic to work.
    """

    # A true pseudo-random number generator in the space of 2 seconds used
    # to add random miliseconds to entries that have same time. This is
    # why we exclude Zero.
    twoSeconds=(
        pandas.Series(range(-999,1000))

        # Exclude 0
        .pipe(lambda s: s[s!=0])

        # Randomize in a deterministic way
        .sample(frac=1,random_state=42)
        .reset_index(drop=True)
    )

    # twoSecondsGen=None # to be redefined later




    def __init__(self, kind, id, cache=None, refresh=False):
        # Setup logging
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        # Suffix for DataCache table
        self.kind=kind

        # ID of dataset
        self.id=id

        # Last updated at
        # self.asof = None

        self.cache=cache
        self.nextRefresh=refresh

        # self.twoSecondsGen=Portfolio.pseudoRandomUniqueMilliseconds()

        # Force data load
        self.balance
        self.ledger



    def getFund(self, subset=None, name=None, currencyExchange=None):
        """
        Given one or more investment items names, passed in the subset
        attribute, return a Fund object which allows handling it as shares
        with share value and currency.
        """

        if subset is None or (isinstance(subset,list) and len(subset)==0):
            # Return all funds
            return Fund(
                name             = name,
                ledger           = self.ledger,
                balance          = self.balance,
                currencyExchange = currencyExchange,
            )
        else:
            if isinstance(subset, str):
                # If only 1 fund passed, turn it into a list
                subset=[subset]


            # And return only this subset of funds
            return Fund(
                name             = name,
                ledger           = self.ledger[self.ledger.fund.isin(subset)],
                balance          = self.balance[self.balance.fund.isin(subset)],
                currencyExchange = currencyExchange,
            )



    def funds(self):
        """
        Return list of tuples as:
        [
            ('fund1',['currency1']),
            ('fund2',['currency1', 'currency2']),
            ('fund3',['currency2'])
        ]

        Which is the list of funds, each with the currencies they present data.
        """
        nonMonetary={'fund', 'time', 'comment'}

        return list(
            self.ledger[
                ['fund'] +
                # Get only currency names
                list(set(self.ledger.columns) - nonMonetary)
            ]

            .set_index('fund')


            # Get rid of completely empty rows
            .dropna(how='all')


            # replace by True where there is NaN
            .isnull()


            # Convert a False (has value) to currency name
            .apply(
                lambda x: x.index[x==False][0],
                axis=1
            )


            # make fund name a regular column
            .reset_index()


            .drop_duplicates()



            # Group by fund and make list of currencies
            .groupby(by='fund', observed=True).agg(list)



            # make fund name a regular column
            .reset_index()



            # convert to tuples
            .itertuples(index=False, name=None)
        )



    @property
    def balance(self):
        return self.getProperty('balance')



    @property
    def ledger(self):
        return self.getProperty('ledger')



    @property
    def asof(self):
        """
        Timestamp of most recent balance or ledger data of this portfolio asset
        """
        return numpy.max(
            [t for t in
                [
                    self.ledger.time.max()  if self.has_ledger  else None,
                    self.balance.time.max() if self.has_balance else None
                ] if t is not None
            ]
        )



    def getProperty(self,prop):
        if getattr(self,f'has_{prop}') is False:
            return None

        if self.nextRefresh:
            self.callRefreshData()
            self.nextRefresh=False
        elif getattr(self,f'_{prop}') is None:
            if self.tryCacheData() is False:
                self.callRefreshData()
        else:
            return getattr(self,f'_{prop}')

        # At this point we have raw data from cache or internet

        # Data cleanup and feature engineering
        self.processData()

        return getattr(self,f'_{prop}')



    def tryCacheData(self):
        ledger_age = None
        balance_age = None

        if self.cache is not None:
            (self._ledger,ledger_age)   = self.cache.get(
                kind=f'{self.kind}__ledger',
                id=self.id
            )

            (self._balance,balance_age) = self.cache.get(
                kind=f'{self.kind}__balance',
                id=self.id
            )

        if ledger_age or balance_age:
            return True

        return False



    def cacheUpdate(self):
        """
        Set new data to DataCache
        """
        if self.cache is not None:
            if self._ledger is not None:
                self.cache.set(kind=f'{self.kind}__ledger', id=self.id, data=self._ledger)
            if self._balance is not None:
                self.cache.set(kind=f'{self.kind}__balance', id=self.id, data=self._balance)



    def callRefreshData(self):
        self.refreshData()
        self.cacheUpdate()



    ############################################################################
    ##
    ## Virtual methods.
    ##
    ## Need to be defined in derived classes.
    ##
    ############################################################################

    def refreshData(self):
        """
        Called when Portfolio feels the need to update its data.
        Pure virtual method, needs to be implemented in derived classes.
        """
        pass



    def processData(self):
        """
        Called right after raw data is loaded from cache or from its original
        source (API) to clean it up.

        Pure virtual method, needs to be implemented in derived classes.
        """
        pass



    @property
    def has_balance(self):
        # Let derived classes define this
        return None



    @property
    def has_ledger(self):
        # Let derived classes define this
        return None



    ############################################################################
    ##
    ## Utility methods, to be used in general situations by derived classes
    ##
    ############################################################################


#     def pseudoRandomUniqueMilliseconds():
#         # Cycle over self.twoSeconds which has 1998 entries (0 to 1997) with
#         # random milliseconds in the range [-999..-1, 1..999]

#         twoSecondsLength=len(Portfolio.twoSeconds)

#         i=0
#         while i<twoSecondsLength:
#             # print('generating')
#             yield pandas.to_timedelta(Portfolio.twoSeconds[i],unit='ms')
#             i+=1
#             if (i==twoSecondsLength):
#                 i=0



    def normalizeTime(time, naiveTimeShift=12*3600) -> pandas.DataFrame:
        """
        Get a pandas.Series in ‘time’ and normalize it:

        1. Add naiveTimeShift (in seconds) to time-naive entries
        2. Add current timezone to TZ-naive entries
        3. Convert all to UTC timezone
        4. De-duplicate timestamps using Portfolio.twoSeconds for small
           adjustments

        Return a normalized pandas.Series ordered and indexed identical to
        input.

        Derived classes must use this method to homogenize time handling
        across the fremework.
        """
        def randomTimedeltas(index):
            """
            Return a vector of random pandas.Timedelta indexed by `index`.
            """
            vec=(
                pandas.to_timedelta(
                    # Create a long random vector by concatenating
                    # Portfolio.twoSeconds multiple times
                    pandas.concat(
                        # How many times we need to duplicate
                        # Portfolio.twoSeconds?
                        max(2,int(
                            numpy.ceil(
                                len(index) /
                                len(Portfolio.twoSeconds)
                            )
                        )) *
                        [Portfolio.twoSeconds]
                    )

                    # Get a vector with precise size
                    .head(len(index)),

                    # This is in milliseconds
                    unit='ms'
                )
            )

            # Overwrite index
            vec.index=index

            return vec

        # Convert naiveTimeShift into something more useful
        timeShift=pandas.to_timedelta(naiveTimeShift, unit='s')

        # Get current timezone
        currtz=(
            datetime.datetime.now(datetime.timezone.utc)
            .astimezone()
            .tzinfo
        )

        instrumented=(
            time

            # Shift dates that have no time (time part is 00:00:00) to the
            # middle of the day (12:00:00) using naiveTimeShift parameter
            .apply(
                lambda cell:
                    pandas.Timestamp(cell+timeShift)
                    if cell.time()==datetime.time(0)
                    else pandas.Timestamp(cell)
            )

            # Convert all to current time zone, or simply add current time
            # zone to TZ-naive entries.
            .apply(
                lambda cell:
                    cell.tz_localize(currtz) # Set TZ on undefined cells
                    if cell.tzinfo is None
                    else cell.tz_convert(currtz) # Convert existing TZ
            )

            # Keep it internally as UTC for more precise and compatible
            # joins.
            .dt
            .tz_convert('UTC')
        )

        # Cirurgically adjust time adding a few random milliseconds only
        # on duplicate items
        dups=instrumented.duplicated()
        duplicated=instrumented[dups]

        instrumented=pandas.concat(
            [
                instrumented[~dups],
                instrumented[dups]+randomTimedeltas(instrumented[dups].index)
            ]
        )

        return instrumented






class PortfolioAggregator(Portfolio):
    def __init__(self, cache=None, refresh=False):
        self.members=[]

        super().__init__(
            kind       = 'aggregator',
            id         = None,
            cache      = cache,
            refresh    = refresh
        )


    def append(self,portfolio):
        if isinstance(portfolio, list):
            self.members += portfolio
        else:
            self.members.append(portfolio)



    ############################################################################
    ##
    ## Virtual methods.
    ##
    ## Need to be defined in derived classes.
    ##
    ############################################################################

    def refreshData(self):
        self.submitToMembers('refreshData')



    def processData(self):
        self.submitToMembers('processData')



    @property
    def has_balance(self):
        has = False
        for p in self.members:
            has = has or p.has_balance
            if has:
                break

        return has



    @property
    def has_ledger(self):
        has = False
        for p in self.members:
            has = has or p.has_ledger
            if has:
                break

        return has



    @property
    def _balance(self):
        return (
            pandas.concat(
                [p._balance for p in self.members if p.has_balance]
            )
            .reset_index(drop=True)
        )



    @property
    def _ledger(self):
        return (
            pandas.concat(
                [p._ledger for p in self.members if p.has_ledger]
            )
            .reset_index(drop=True)
        )



    @property
    def asof(self):
        return numpy.max([p.asof.tz_localize(None) for p in self.members])



    def tryCacheData(self):
        self.submitToMembers('tryCacheData')



    def submitToMembers(self, method: str):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            tasks={}
            for p in self.members:
                m = getattr(p,method)
                task[executor.submit(m)] = p

            for task in concurrent.futures.as_completed(tasks):
                self.debug(f'Done {tasks[task]}')

                # The return of submited method (processData) or the raise of error
                task.result()



    def __repr__(self):
        return '{klass}({members})'.format(
            klass   = type(self).__name__,
            members = ', '.join([m.__repr__() for m in self.members])
        )



    def to_markdown(self, title_prefix=None):
        title="{klass}({n} members)".format(
            klass=type(self).__name__,
            n=len(self.members)
        )

        body=[
            (m.to_markdown(title_prefix=title_prefix + '#' if title_prefix else None))
            for m in self.members
        ]

        if title_prefix is None:
            return (title,body)
        else:
            body='\n\n'.join(body)
            return f"{title_prefix} {title}\n{body}"


